//!The RAq Approach: matrix.h
/*!
Generates the distance matrix necessary for the creation of tree.
Using the desired metric as a parameter, calculates distances according
to that method and calculates respective distance matrix.
*/
//(c) 2006 Suzanne Matthews
/*
This file is part of the RaqApproach software package.

    RaqApproach is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    RaqApproach is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with RaqApproach; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef MATRIX_H_
#define MATRIX_H_

#include <math.h>
#include "elements.h"
#include <cstdlib>
#include <cstdio>
#include "common.h"
#include <algorithm>

using namespace std;

int find_length(string one, string two); //!< Returns the length of the shortest of the two strings
double calculateDistance(string first, string last, int dist_type, const int penalty); //!< Calculates distance between two strings
double uncorrected(string one, string two, const int penalty); //!< Calculates the uncorrected distance
double jukes_cantor(string one, string two, const int penalty);//!< Calculates the JC distance (needs to be worked on)
double kimura(string one, string two);//!< Calculates the K2P distance (needs to be worked on)
int calculateMax(int max_score); //!< Returns the value of the maximum score: for score-based distance method
double calculateScore(string first, string second); //!< Calculates the score for score-based distance method
bool is_transition(char achar, char bchar); //!< Determines if chars are transitions or transversions


void init_unknowns()
{
	unknown_chars.push_back('S');
	unknown_chars.push_back('W');
	unknown_chars.push_back('H');
	unknown_chars.push_back('B');
	unknown_chars.push_back('V');
	unknown_chars.push_back('D');
	unknown_chars.push_back('N');
	unknown_chars.push_back('?');

}

void preprocess(string & newstring ) { //!< we assume that the input is correct
	for ( unsigned int i = 0; i < newstring.size(); ++i ) {
		char temp = toupper(newstring[i]);
		if ( (temp != 'A') &&
		     (temp != 'G') &&
		     (temp != 'C') &&
		     (temp != 'T') &&
		     (temp != 'R') &&
		     (temp != 'Y') &&
		     (temp != '-') )
		     newstring[i] = unknown_char; //!< even errors will be processed as a '?'
	}
}
multimap<double, string> create_matrix(int dist_type, const int penalty) {

  //!input: text file
  ifstream fin;
  fin.open("taxa.txt");
  if (!fin) { //!< if text file cannot be opened, return error_matrix; exit gracefully
    cout << "file cannot be opened." << endl;
    multimap<double, string> error_matrix;
    error_matrix.insert(make_pair(0.0, "error"));
    return error_matrix;
  }

   //!input into array
  string newstring;
  string numtaxa_string;
  multimap<double, string> matrix;
  list<string> taxa;
  cout << "preprocessing taxa file..." << endl;
  fin >> numtaxa_string;
  stringstream sst;
  sst << numtaxa_string;
  sst >> num_taxa;
  cout << "num_taxa is: " << num_taxa << endl;
  while ( !fin.eof() ){ //!< while not at end of file
    fin >> newstring;
    preprocess(newstring); //!< preprocess string (operation: O(nl))
    taxa.push_back(newstring); //!< input taxa into array
  }
  cout << "done." << endl;

  fin.close(); //!< close file
  list<string>::iterator istr = taxa.begin();
  if ( DISPLAY ) {
	  cout << "outputting processed taxa:" << endl;
	  for (istr; istr != taxa.end(); ++istr) { //!< print taxa to standard output
		cout << (*istr) << endl;
	  }
  }

  istr = taxa.begin();
  list<string>::iterator istr2 = taxa.begin();
  cout << "calculating distance matrix..." << endl;
  float dist = 0.0;
  int i,j;
  i = j = 1;
  for ( istr; istr !=  taxa.end(); ++istr) {
    istr2 = istr;
    ++istr2;
    for ( istr2; istr2 != taxa.end(); ++istr2) {
      ++j;
      dist = calculateDistance((*istr), (*istr2), dist_type, penalty); //!< calculate distance between all pairs of taxa (n choose 2)
      if ( DISPLAY ) {
		  cout << "distance between taxa " << i << " and " << j;
		  cout << " is: " << dist << endl; //!< print calculated distance to standard output
	  }
      stringstream ss, ss2;
      string str;
      string str2;
      ss << i;
      ss >> str;
      ss2 << j;
      ss2 >> str2;
      string tag = str + "." + str2; //!< create distance marker (A.B) for taxa A and B
      matrix.insert( make_pair(dist, tag) ); //!< insert newly calculated distance into matrix
    }
    ++i;
    j = i;
  }
  cout << "distance matrix creation complete." << endl;
  if ( DISPLAY ) {
	  multimap<double, string>::iterator m_itr = matrix.begin();
	  while ( m_itr != matrix.end() ) {
		cout << m_itr ->first << " " << m_itr->second << endl; //!< print contents of distance matrix to standard output
		++m_itr;
	  }
  }
  return matrix;

}

int calculateMax(int max_score) { //!< calculates the max score for the score-based distance method. currently a dummy function.
  int newmax = 0;
  newmax = max_score; //!< note how it's currently only returning its parameter
  return newmax;
}

double calculateDistance(string one, string two, int dist_type, const int penalty) { //!< Calculates distance matrix

  double distance = 0;

  switch(dist_type) {
  case 1: distance = uncorrected(one, two, penalty); //!< calculates uncorrected distance (requires a gap penalty)
    break;
  case 2: distance = jukes_cantor(one, two, penalty); //!< calculates jukes_cantor distance (requires a gap penalty)
    break;
  case 3: distance = kimura(one, two); //!< calculates kimura distance ( does not require a gap penalty)
    break;
  case 4: distance = calculateScore(one, two); //!< calculates score-based distance (gap penalty set internally)
    break;
  default: cout << "invalid option for distance!" << endl;
  }

  return distance;
}

int find_length(string one, string two) { //!< returns length of shortest string

  int length = 0;
  if ( one.size() < two.size() )
    length = one.size();
  else
    length = two.size();

  return length;
}

double uncorrected(string one,string two, const int gapp) { //!< calculates the uncorrected distance


  double distance = 0.0;
  int matches = 0;
  int positions_scored = find_length(one, two);

  int gaps = 0;

  for ( unsigned int i = 0; i < positions_scored; ++i ) {
    if ( (one[i] == two[i]) && (one[i]!= unknown_char) ) {
      matches++;
    }
    else {
      if ( one[i] == gap_char || two[i] == gap_char )
		gaps++;
    }
  }

  double temp = gaps*gapp; //!< multiply...
  temp = temp + positions_scored;
  double S = matches / temp; //!< and divide to yield S
  distance = 1 - S; //!< distance = 1 - S
  return distance;
}

double jukes_cantor(string one, string two, const int gapp) { //!< calculates the jukes cantor distance (does not work right now)
  double distance = 0.0;
  const float b = .75; //!< b is 3/4 for nucleic acids... 19/20 for amino acids

  double D = uncorrected(one, two, gapp); //!< first calculates the uncorrected distance
  if ( D >= b) //!< if it's greater that D (we run into issues... so we are just returning D)
    return D;
  D = (double)D/(double)b; //!< else, we divide
  D = 1 - D;
  distance = log(D); //!< preliminary distance is: log(1-D)
  distance = distance*b;
  distance = distance*-1; //!< final distance = -b*ln(1-(D/b))

  return distance;
}

bool matches_unknowns(char a_char) { //!< checks to see if it is an unknown character
	vector<char>::iterator v_itr = unknown_chars.begin();
	v_itr = find(unknown_chars.begin(), unknown_chars.end(), a_char);
	if ( v_itr != unknown_chars.end() )
		return true;
	return false;
}

double kimura(string one, string two) { //!< calculates kimura distance: uses K2P
  int positions_scored = find_length(one, two);
  int transitions = 0;
  int transversions = 0;
  double distance = 0.0;

  for ( unsigned int i = 0; i < positions_scored; ++i) {
    if ( matches_unknowns(one[i]) || one[i]== gap_char ) //!< ignore unknown_char and gap_char
      continue;
    else if ( matches_unknowns(two[i]) || two[i] == gap_char ) //!< ignore unknown_char and gap_char
      continue;
    else if ( is_transition(one[i],two[i]) ) //!< if it is a transition
      ++transitions; //!< increment counter
    else
      ++transversions; //!< else, increment transversion counter
  }

  double P = (double)transitions / (double)positions_scored; //!< calculate P: transitions/positions_scored
  double Q = (double)transversions / (double)positions_scored; //!< calculate Q: transversions/positions_scored

  double temp =  2*Q; //!<  2Q
  temp = 1 - temp; //!<  (1 - 2Q)
  temp = sqrt(temp); //!<  sqrt(1-2Q)
  double newtemp = 2*P; //!<  2P
  newtemp = newtemp - Q; //!<  (2P - Q)
  newtemp = 1 - newtemp; //!<  (1 - 2P - Q)
  temp = newtemp*temp; //!<  (1-2P-Q)*sqrt(1-2Q)
  temp = log(temp); //!<  ln[(1-2P-Q)*sqrt(1-2Q)]
  distance = -.5 * temp; //!<  -.5 * ln[(1 - 2P - Q)*sqrt(1-2Q)]

  return distance;
}


double calculateScore(string one, string two) { //!< calculates the score-based distance (a novel method based on K2P)
  double score=0.0;
  //!the following values are constants
  //!they can be adjusted as necessary
  //**********************************
  const int match_score = 1;
  const float alpha = .5;
  const float beta = .25;
  //*********************************
  init_unknowns();
  bool penalize_end_gaps = true; //!< this value can also be modified; by default, it is true.

  int max_score = 0;

  max_score = find_length(one, two);
  max_score = calculateMax(max_score); //!< calculate the max score

  int gapscore;
  if ( penalize_end_gaps == true ) //!< if we should penalize end gaps
    gapscore = -1;
  else
    gapscore = 0;

  unsigned int i = one.size()-1;
  unsigned int j = two.size()-1;

  for (unsigned int x = 0; x < one.size(); ++x) { //!< calculate the score
    if ( (one[x] == two[x]) ) { //!< if they match
      if ( ( one[x] != gap_char) )
      	score++; //!< add one
      else
      	score = score + 0; //!< matching gaps get a score of +0
    }
    else {
      if (one[x] == gap_char || two[x] == gap_char) { //!< if gap, subtract one
		score--;
      }
      else if ( (one[x] == unknown_char) || (two[x] == unknown_char) )
      	score = score + 0;
      else if (is_transition(one[x], two[x]) ) { //!< if it is a transition
		score = score+alpha; //!< add alpha score
      }
      else {
		score = score+beta; //!< else, add beta score.. even errors will get logged as this
      }
    }
  }
  double distance = (max_score - score)/max_score; //!< normalized distance = (Max score - observed score)/max score
  return distance;
}

bool is_transition( char achar, char bchar ) { //!< calculates if inputted characters are transversions or transitions
  //!< returns true if it is a transition, false if it is a transversion
  /*Note: A-G and C-T are transitions. All else are transversions*/

  char newa = toupper(achar);
  char newb = toupper(bchar);

  switch ( newa ) {

  case 'A': {
    if ( ( newb == 'G') || (newb == 'R') )
      return true;
    else if (newb == 'A')
      return true;
    else return false;
  }
    break;
  case 'G': {
    if ( (newb == 'A') || ( newb == 'R') )
      return true;
    else if ( newb == 'G')
      return true;
    else return false;
  }
    break;
  case 'C': {
    if ( ( newb == 'T') || ( newb == 'Y') )
      return true;
    else if ( newb =='C')
      return true;
    else return false;
  }
    break;
  case 'T': {
    if ( (newb == 'C') || (newb == 'Y') )
      return true;
    else if ( newb == 'T')
      return true;
    else return false;
  }
    break;
  case 'R': {
	if ( (newb == 'A') || (newb == 'G') )
	  return true;
	else if ( newb == 'R')
	  return true;
	else return false;
  }
    break;
  case 'Y': {
	if ( (newb == 'C') || (newb == 'T') )
	  return true;
	else if ( newb == 'Y')
	  return true;
	else return false;
  }
    break;
  default: cout << "ERROR! Not valid DNA sequence!: " << newa << endl;
  }
}

#endif
